/*
 * Copyright (c) 1996, 1997 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 */

#include "iiop-link.h"

#include <signal.h>

/* A signal handler is a pointer to function taking int returning void. */
typedef void (*signal_handler_t)(int);

/*****************************************************************************/

static void printbuf(FLICK_BUFFER *buf)
{
	int i;
	int size = buf->buf_read - buf->real_buf_start;
	printf("Buffer size: %06x", size);
	for (i = 0; i < size; i++) {
		if (!(i & 15))
			printf("\n%06x:", i);
		if (!(i & 7))
			printf(" ");
		printf("%02x ", ((int)((char *)buf->real_buf_start)[i]) & 0xFF);
	}
	for (i = 0; i < size; i++) {
		char c = ((char *)buf->real_buf_start)[i];
		if ((c & 128) ||
		    !(c & ~31))
			c = '?';
		if (!(i & 15))
			printf("\n%06x:", i);
		if (!(i & 7))
			printf(" ");
		printf(" %c ", c);
	}
	printf("\n");
}

static int recv_buf(int sock, FLICK_BUFFER *out_buf)
{
	/* Semantics of this call:
	   This call should block until it gets a message.  Reading a
	   message: on the first read, the first word received will be a
	   network order byte count, including the first word.  all
	   remaining reads should block until the entire message is
	   received.  */

	int real_buf_size = out_buf->real_buf_end - out_buf->real_buf_start;
	int bytes_read = 1; /* init to non-zero value */
#ifdef TRACK_FRAGMENTATION
	int mult = 0;
#endif
	if ((out_buf->buf_read == out_buf->real_buf_start)/* new buffer */
	    || (out_buf->buf_end >= out_buf->buf_read)) /* buf is empty */ {
		out_buf->buf_start = out_buf->buf_read = out_buf->real_buf_start;
		
		if ((bytes_read = read(sock, (char *)out_buf->buf_read,
				       real_buf_size)) > 0)
			out_buf->buf_read += bytes_read;
		
	} else {
		/* go to the next message */
		out_buf->buf_start = out_buf->buf_end;
	}
	
	if (bytes_read > 0) {
		
		int test = 1;
		int buf_size = ((unsigned long *)out_buf->buf_start)[2];
		if (((char *)out_buf->buf_start)[6] != ((char *)&test)[0])
			buf_size = ((buf_size >> 24) & 0x000000FF) |
				   ((buf_size >>  8) & 0x0000FF00) |
				   ((buf_size <<  8) & 0x00FF0000) |
				   ((buf_size << 24) & 0xFF000000);
		buf_size += 12; /* Offset for the header... */
		
		out_buf->buf_end = out_buf->buf_start + buf_size;
		
		/* if only partial message received */
		if (out_buf->buf_end > out_buf->buf_read) {
			
			/* if the rest of the message won't fit */
			if (out_buf->buf_end > out_buf->real_buf_end) {
				
				/* copy the partial message into the
                                   beginning of the buffer */
				memcpy(out_buf->buf_start, out_buf->real_buf_start,
				       out_buf->buf_read - out_buf->buf_start);
				/* adjust buf_read */
				out_buf->buf_read -= out_buf->buf_start
						     - out_buf->real_buf_start;
				
				/* if the message is bigger than the buffer */
				if (buf_size > real_buf_size) {
					/* buf_start used as storage for
                                           out_buf->real_buf_start */
					out_buf->buf_start = t_realloc
							     (out_buf->
							      real_buf_start, char,
							      buf_size);
					
					/* adjust real_buf and buf values
                                           because of realloc */
					real_buf_size = buf_size;
					out_buf->real_buf_end =/* same as buf_end */
					out_buf->buf_end = out_buf->buf_start
							  + buf_size;
					out_buf->buf_read +=
						(out_buf->buf_start
						 - out_buf->real_buf_start);
					out_buf->real_buf_start =
						out_buf->buf_start;
				} else {
					out_buf->buf_start =
						out_buf->real_buf_start;
					out_buf->buf_end = out_buf->buf_start
							   + buf_size;
				}
			}
			do {
#ifdef TRACK_FRAGMENTATION
				mult++;
#endif
				if ((bytes_read = read(sock, (char *)out_buf
						       ->buf_read,
						       real_buf_size
						       - (out_buf->buf_read
							  - out_buf->real_buf_start)
					)) > 0)
					out_buf->buf_read += bytes_read;
			} /* until the whole message is read */
			while (bytes_read > 0 && out_buf->buf_end
			                         > out_buf->buf_read);
		}
		
	}
#ifdef TRACK_FRAGMENTATION
	if (mult)
		printf("Multiple (%d) reads on server!\n", mult);
#endif
	/* printbuf(out_buf); */
	if (bytes_read <= 0) {
		close(sock);
		return 0;
	}
	
	return 1;
}

static int send_buf(int sock, FLICK_BUFFER *in_buf)
{
	signal_handler_t saved_sigpipe_handler;
	
	int size = (char *)in_buf->buf_current - (char *)in_buf->buf_start;
	int written = 0;
	int offset = 0;
#ifdef TRACK_FRAGMENTATION
	int mult = 0;
#endif
	/* in_buf->buf_read = in_buf->buf_current; printbuf(in_buf); */
	
	/* Ignore SIGPIPE: writes on a socket with no reader. */
	saved_sigpipe_handler = signal(SIGPIPE, SIG_IGN);
	
	/* I'm not sure if this while loop is necessary... */
	((unsigned long *)in_buf->buf_start)[2] = size - 12;
	while (size > 0) {
		int amount2write = (size < PACKET_SIZE) ? size : PACKET_SIZE;
		written = write(sock,
				&((char *)in_buf->buf_start)[offset],
				amount2write);
#ifdef TRACK_FRAGMENTATION
		if (offset)
			mult++;
#endif
		if (written < 0)
			break;
		
		size -= written;
		offset += written;
	}
	
	/* Restore the old SIGPIPE handler. */
	(void) signal(SIGPIPE, saved_sigpipe_handler);
	
#ifdef TRACK_FRAGMENTATION
	if (mult)
		printf("Multiple reads on server - %d!\n", mult);
#endif
	if (size > 0)
		perror("socket write failure");
	
	/* Return 1 (true) for success or 0 (false) for failure. */
	return (size == 0);
}


/*****************************************************************************/

int flick_buffer_contains_another_whole_message(FLICK_BUFFER *out_buf) {
#define next_buf_start (out_buf->buf_end)
	
	if (next_buf_start < out_buf->buf_read) /* buf is not empty */ {
		int test = 1;
		int buf_size = ((unsigned long *)next_buf_start)[2];
		if (((char *)next_buf_start)[6] != ((char *)&test)[0])
			buf_size = ((buf_size >> 24) & 0x000000FF) |
				   ((buf_size >>  8) & 0x0000FF00) |
				   ((buf_size <<  8) & 0x00FF0000) |
				   ((buf_size << 24) & 0xFF000000);
		buf_size += 12; /* Offset for the header... */
		
		/* if whole message found */
		return (next_buf_start + buf_size <= out_buf->buf_read);
	} else {
		return 0; /* return false */
	}
#undef next_buf_start
}


/*****************************************************************************/

int flick_client_send_request(FLICK_CLIENT in_cli, FLICK_BUFFER *in_buf)
{
	return send_buf(in_cli->boa->socket_fd, in_buf);
}

int flick_client_get_reply(FLICK_CLIENT in_cli, FLICK_BUFFER *out_buf)
{
	return recv_buf(in_cli->boa->socket_fd, out_buf);
}

int flick_server_get_request(int socket, FLICK_BUFFER *out_buf)
{
	return recv_buf(socket, out_buf);
}

int flick_server_send_reply(int sock, FLICK_BUFFER *in_buf)
{
	return send_buf(sock, in_buf);
}

int flick_server_send_exception(int sock,
				unsigned long request_id,
				int exception_type)
{
	FLICK_BUFFER flick_buf;
	
	char buf[(  12 /* A `GIOP::MessageHeader'/ */
		  + 12 /* Our `GIOP::ReplyHeader'/ */
		  + 12 /* Our system exception.    */
		)];
	
	/*
	 * Manufacture the `GIOP::MessageHeader'.  See Section 12.4.1 of the
	 * CORBA 2.0 specification.
	 */
	buf[0] = 'G';
	buf[1] = 'I';
	buf[2] = 'O';
	buf[3] = 'P';
	buf[4] = 1;
	buf[5] = 0;
	buf[6] = flick_is_little_endian;
	buf[7] = 1;
	
	*((unsigned int *) &buf[8])
		= (12  /* Our `GIOP::ReplyHeader' (see below). */
		   + 4 /* The exception identifier (int, should be string). */
		   + 8 /* The system exception body. */
			);
	
	/*
	 * Manufacture the `GIOP::ReplyHeader'. See Section 12.4.2.
	 */
	*((unsigned int *) &buf[12]) = 0;
	*((unsigned int *) &buf[16]) = request_id;
	*((unsigned int *) &buf[20]) = CORBA_SYSTEM_EXCEPTION;
	
	/*
	 * Encode the system exception.  XXX --- The exception type should be a
	 * string, but we use integers for now.
	 */
	*((int *) &buf[24]) = exception_type;
	*((unsigned int *) &buf[28]) = 0;
	*((unsigned int *) &buf[32]) = CORBA_COMPLETED_NO;
	
	/* Send this exceptional reply back to the client. */
	flick_buf.buf_start = buf;
	flick_buf.buf_current = buf + 36;
	flick_buf.buf_end = buf + 36;
	
	return send_buf(sock, &flick_buf);
}

/* End of file. */

